<?php

// SrcDS Connectivity Checker
// Convenience tool that checks if TCP and UDP are reachable at the given IP address and port number

// Query strings:
// ip - default: the client's IPv4 address. This only works if the site is loaded over IPv4.
// port - default: 27015 - required in API mode
// runtest - if this is present, run server test (webpage mode only)

// Also:
// If the HOST header begins with "api.", this will output a text-only response so that it can be used programmatically
// i.e. if you host a site with "api.example.com" and that's how this is called, it will output one of the following:
// unreachable, tcponly, udponly, tcpudp
//
// Otherwise, it outputs a webpage that can be used in a browser.

require __DIR__ . '/settings.php';
require __DIR__ . '/ratelimiter.php';
require __DIR__ . '/SourceQuery/bootstrap.php';

use xPaw\SourceQuery\SourceQuery;

function validPort($port)
{

	if( is_null($port) )
		return FALSE;

	if( $port < 1023 || $port > 65535 || !ctype_digit($port) )
	{

		return FALSE;

	}
	else
	{

		return TRUE;

	}

}

function validIpv4($ip)
{

	if( is_null($ip) )
		return FALSE;

	# Separate each octet of the IP address into an array
	$temp_ip = explode('.', $ip);

	# check that the resulting array is 4 octets in size
	if( count($temp_ip) != 4 )
	{
		return FALSE;
	}

	# Accumulator (array index)
	$index = 0;

	# Process each octet
	foreach ($temp_ip as $octet) {

		# Check that all of octet is numeric
		if( !ctype_digit($octet) )
			return FALSE;

		# Check that octet value is between 0 and 255
		if( $octet < 0 || $octet > 255 )
			return FALSE;

		$index += 1;

	}

	return TRUE;

}

// Gets server info. Returns server info array, or FALSE if error.
function get_server_info(string $ip, int $port)
{

	static $serverInfo = NULL;

	// If this function has already run, return the last result
	// Prevents repeatedly querying the server
	if( !is_null($serverInfo) )
		return $serverInfo;

	$Query = new SourceQuery( );
	
	try
	{
		$Query->Connect( $ip, $port, 1, SourceQuery::SOURCE );
		$serverInfo = $Query->GetInfo();
	}
	catch( Exception $e )
	{
		$serverInfo = FALSE;
		return FALSE;
	}
	finally
	{
		$Query->Disconnect();
	}

	return $serverInfo;

}

function check(string $ip, int $port)
{

	// Status will contain the result of connectivity checks
	// 0 = unreachable, 1 = tcponly, 2 = udponly, 3 = tcpudp
	$status = 0;

	// tcp connection check
	$tcp_conn = @fsockopen("tcp://".$ip, $port, $error_code, $error_message, 1);

    if (is_resource($tcp_conn))
    {

    	fclose($tcp_conn);
    	$status += 1;

    }

    // udp connection check
    if( get_server_info($ip, $port) )
    	$status += 2;

    // return
    switch ($status) {
    	default:
    		return "unreachable";
    	case 1:
    		return "tcponly";
    	case 2:
    		return "udponly";
    	case 3:
    		return "tcpudp";
    }

}

function main()
{

	# Don't report notices or other trivial errors if DEBUG_MODE is FALSE
  if( !DEBUG_MODE )
  {

    $DISABLED_MESSAGES =
      E_NOTICE|E_USER_NOTICE|E_STRICT|E_DEPRECATED|E_USER_DEPRECATED;

    error_reporting( error_reporting() - $DISABLED_MESSAGES );

  }

	header("Cache-Control: private, no-store, no-cache, must-revalidate, proxy-revalidate, no-transform");

	if( !db_conn() )
	{
		error_log("Database connection error. Cannot continue...");
		return 500;
	}

	// Initialize the IP address as the IP of the requestor
	$ip = $_SERVER['REMOTE_ADDR'];

	if( isset($_GET['port']) && validPort( trim($_GET['port'])) )
	{
		$port = trim($_GET['port']);
	}
	else
	{
		$port = 27015;
	}

	// If the URI begins with "/api", give a text-only response
	// This was intended to be used with Apache rewrite rules to rewrite /api to /
	if( substr( $_SERVER['REQUEST_URI'], 0, 4) == "/api" )
	{

		header("Content-Type: text/plain");
		header("X-Robots-Tag: noindex");

		if( !validIpv4($ip) )
		{

			echo "ERROR: Invalid IPv4 address. You must access this service over IPv4.";
			return 400;

		}

		// Although we already have the port set, check the query string here
		// If it's not explicitly set, bail out so we don't trigger from
		// spam bots/web crawlers/etc
		if( !validPort($port) || is_null($_GET['port']) || $_GET['port'] < 1024 )
		{

			echo "ERROR: Invalid port number. Query string 'port' must be set to a valid port number greater than 1023.";
			return 400;

		}

		if( is_ratelimited() )
		{

			echo "ratelimited";
			return 429;

		}

		ratelimit_counter_add();

		echo check($ip, $port);
		return(200);

	}

	// api mode response end

	// If user is accessing us directly, redirect to the intended webpage
	if($_GET['src'] != "iframe")
	{
		header("Location: " . REDIRECT_TO_ADDR);
		exit();
	}

	$portInputValue = "27015";

	// pre-fill the input box using previous form data
	if( isset($port) )
		$portInputValue = $port;

	// Run the server test
	if( isset($_GET['runtest']) )
	{

		if( is_ratelimited() )
		{

			$serverReachability = "ratelimited";

		} else {

			ratelimit_counter_add();

		}

		if( validIpv4($ip) && validPort ($port) && !isset($serverReachability) )
		{
			$serverReachability = check($ip, $port);
		}
		elseif ( !isset($serverReachability) )
		{
			$serverReachability = "error";
		}

		$resultH1 = "Result:";

		switch ($serverReachability)
		{
			default:
				$resultTitle = "<test-fail>ERROR</test-fail>";
				break;

			case "tcpudp":
				$resultTitle = "<test-pass>PASS</test-pass>";
				break;

			case "tcponly":
				$resultTitle = "<test-fail>FAIL</test-fail>";
				break;

			case "udponly":
				$resultTitle = "<test-warning>PASS*</test-warning>";
				break;

			case "unreachable":
				$resultTitle = "<test-fail>FAIL</test-fail>";
				break;
		}

	} else {

		$resultH1 = "";
		$resultTitle ="";

	}

	?>

	<!DOCTYPE html>
	<html lang="en">
		<head>
			<meta charset="UTF-8">
			<title>SrcDS Connectivity Test</title>
			<link rel="stylesheet" href="styles.css">

			<script>

				function resultInProgress()
				{
					var result = document.getElementById("result");
					result.innerHTML = "TESTING...";

					var resultInfo = document.getElementById("result-info");
					resultInfo.innerHTML = "";

					document.getElementById("submit").disabled = true;

				}
		
			</script>

		</head>

		<body>

			<form action="/" onsubmit="return resultInProgress();">

				<input type="hidden" name="src" value="iframe">
				<input type="hidden" name="runtest" value="">

				Using your detected IP address
				(<a href="#" onClick="alert('<?= $ip ?>')">show</a>)

				<p>
					Port:
					<input style="width: 6em" type="number" name="port" min="1024" max="65535" placeholder="27015" value="<?= $portInputValue ?>">
				</p>

				<p>The default server port is 27015.</p>
				<p><input id="submit" type="submit" value="Check Server"></p>

			</form>

			<?php
				if( isset($resultH1) && $resultH1 != "" )
					echo "<hr>";
			?>

			<h3><?= $resultH1 ?></h3>

			<h2 id="result"><?= $resultTitle ?></h2>

			<div id="result-info">

			<?php

				if( isset($serverReachability) )
				{

					switch($serverReachability)
					{
						case "error":
							echo "<h4>An error occurred. Check that you filled out the form correctly.</h4>";
							break;

						case "tcpudp":
							echo "<h4>The server is reachable from the Internet.</h4>";
							break;

						case "tcponly";
							echo "<h4>The TCP port is open, but the server could not be queried over its UDP port.</h4>

								<p>Players will NOT be able to connect to your server.</p>

								<p>Likely causes:</p>

								<ul>
									<li>A firewall is blocking the connection</li>
									<li>You properly forwarded the TCP port but not the UDP port</li>
									<li>Another service on the host machine is using this port, and the dedicated server automatically selected another port number</li>
								</ul>
							";
							break;

						case "udponly":
							echo "
								<h4>The server could be queried over its UDP port, but the TCP port is not open.</h4>
								
								<p>Players will be able to connect to your server but RCON (server remote control) will not work. This is actually more secure if you don't intend to use the built-in RCON feature.</p>

								<p>Likely causes:</p>

								<ul>
									<li>A firewall is blocking the connection</li>
									<li>You properly forwarded the UDP port but not the TCP port</li>
									<li>You intentionally didn't forward/allow the TCP port for security purposes</li>
								</ul>
							";
							break;

						case "unreachable":
							echo "<h4>The server could not be reached.</h4>";

							echo "<p>Likely causes:</p>";
							echo "
								<ul>
									<li>The server is not running</li>
									<li>The connection is blocked by a firewall
										<ul>
											<li>Make sure the ports are allowed through both network and host firewalls (i.e. Windows Firewall)</li>
										</ul>
									</li>
									<li>You are using NAT and your ports are not forwarded
										<ul>
											<li>If the machine hosting the server is using DHCP, the IP may have changed and the ports are no longer forwarded to the correct IP
												<ul>
													<li>Use a Static IP or a DHCP reservation to resolve this</li>
												</ul>
											</li>
											<li>If your ISP uses CGNAT, you will not be able to run a server</li>
										</ul>
									</li>
								</ul>
							";
							break;

						case "ratelimited":
							echo "<h4>You are currently ratelimited. Please try again later.</h4>";
							break;

					}

				}

				?>

			</div>

		</body>

	</html> 

	<?php

	return 200;

}

// Set the HTTP response code to the output of main()
$http_response = main();
if( is_numeric($http_response) && $http_response > 100 && $http_response < 511 )
{

  http_response_code($http_response);

}
else
{

  http_response_code(500);
  error_log('Received an invalid response code from main function.', 0);

}

# return the value of http_response_code to the OS/server
return(http_response_code());